#ifndef WIRECELL_PLANEDUCTOR
#define WIRECELL_PLANEDUCTOR

#include "WireCellIface/IPlaneDuctor.h"
#include "WireCellIface/IDiffusion.h"
#include "WireCellIface/IConfigurable.h"

#include <deque>

namespace WireCell {

    class BufferedHistogram2D;
    class Diffuser;

    // IConverter<IDiffusion::pointer, IPlaneSlice::pointer>
    class PlaneDuctor : public IPlaneDuctor, public IConfigurable {
    public:


	/** Create a PlaneDuctor that produces plane slices from diffusions.
	 *
	 * Note, this class does not apply any field nor electronics
	 * response functions.  It provides a kind of "truth" measure
	 * as if all drifted and diffused charge is perfectly
	 * "collected" by a wire from each plane.  For a more
	 * realistic IPlaneDuctor see the class SignalInducer.
	 *
	 * \param wpid is the wire plane ID given to the resulting
	 * plane slice.
	 *
	 * \param nwires gives the number of wires in the plane.
	 *
	 * \param lbin is the longitudinal (time) bin width.
	 *
	 * \param tbin is the transverse (pitch) bin width.
	 *
	 * \param lpos0 is the longitudinal (time) origin (starting
	 * time of zero'th latch).  Time bins are assumed to cover one
	 * tick, inclusive of the tick start time and exclusive of the
	 * tick end time.
	 * 
	 * \param tpos0 is the transverse (pitch) origin (location of
	 * wire zero along the pitch direction).  Transverse bins are
	 * assumed to be centered on the wire with the low/high edges
	 * extending to a half pitch distance below/above the wire
	 * inclusive/exclusive, respectively.
	 * 
	 * Note: the origin and bin widths should be sync'ed,
	 * quantitatively and conceptually to be the same as used to
	 * form the input diffusion patches.
	 */
	PlaneDuctor(WirePlaneId wpid = WirePlaneId(kUnknownLayer),
		    int nwires = 100,
		    double lbin = 0.5*units::microsecond,
		    double tbin = 5.0*units::millimeter,
		    double lpos0 = 0.0*units::microsecond,
		    double tpos0 = 0.0*units::millimeter);



	virtual ~PlaneDuctor();

	virtual void reset();
	virtual bool operator()(const input_pointer& diffusion, output_queue& outq);

	virtual void configure(const WireCell::Configuration& config);
	virtual WireCell::Configuration default_configuration() const;

	// internal

	// remove any diffusions that are fully past the planeductor's "now"
	void purge_bygone();
	// produce a plane slice from all diffusions that overlap "now"
	IPlaneSlice::pointer latch_hits();
	// purge + latch
	void latch_one(output_queue& outq);
	// return true if we have not buffered enough to assure we see
	// all diffusion patches for the current "now" time.
	bool starved();

    private:

	WirePlaneId m_wpid;
	int m_nwires;
	double m_lbin, m_tbin, m_lpos;
	double m_tpos0;

	// internal buffering
	IDiffusionSet m_input;
	bool m_eos;
    };

}

#endif
